/**
 * Alibaba.com Inc.
 * Copyright (c) 2004-2016 All Rights Reserved.
 */
package com.aliyun.iot;

import com.aliyun.iot.connect.ConnectOpts;
import com.aliyun.iot.internal.Session;
import com.aliyun.iot.internal.SynRespToken;
import com.aliyun.iot.message.AbstractMessage;
import com.aliyun.iot.message.PubAck;
import com.aliyun.iot.message.Publish;
import com.aliyun.iot.message.Request;
import com.aliyun.iot.message.Response;

/**
 * ccp客户端
 *
 *
 * @version $Id: Session.java,v 0.1 2016年1月21日 下午2:25:45 Exp $
 */
public class CcpClient {

    //ccp 的内部会话 ，一旦调用连接后具备自动生成随机密钥、授权、重连、证书更新功能
    private Session ccpSession = null;

    /**
     * construct
     * ccp连接参数，
     *    <b>控制台参数</b>：
     *      productAppKey 产品证书的appkey，通过控制台查看
     *      productAppSecret 产品证书的私钥，通过控制台查看
     *      deviceId     设备证书id，通过控制台查看
     *      deviceSecret 设备证书私钥，通过控制台查看
     *    <b>其它参数</b>
     *      aliveSeconds 心跳时间，最小60秒
     * 
     * @param connectOpts
     */
    public CcpClient(ConnectOpts connectOpts) {
        //check
        if (connectOpts == null) {
            throw new CcpException("connectOpts is null");
        }

        if (connectOpts.getProductAppKey() == null) {
            throw new CcpException("connectOpts.getProductAppKey() is null");
        }

        if (connectOpts.getProductAppSecret() == null) {
            throw new CcpException("connectOpts.getProductAppSecret() is null");
        }

        if (connectOpts.getDeviceId() == null && connectOpts.getDeviceName() == null) {
            throw new CcpException("connectOpts.getDeviceId() or getDeviceName() is null");
        }

        if (connectOpts.getDeviceSecret() == null) {
            throw new CcpException("connectOpts.getDeviceSecret() is null");
        }

        ccpSession = new Session(connectOpts);
    }

    /**
     * ccp session 在引用前应该初始化连接
     * 连接初始化
     * @throws Exception
     */
    public void connect() throws CcpException {
        try {
            this.ccpSession.connect();
        } catch (Exception e) {
            throw new CcpException("连接失败：" + e.getMessage(), e);
        }
    }

    /**
     * 增加监听客户端,异步消息会回调到listener 比如：
     *  publish
     *  unknowsession
     *  ping
     *  以及ack回执消息，注意，如果通过 CcpClient.sendMessageAndWait()同步方式调用，
     *  ack消息将不再回调通知，而是同步返回。
     * @param listener
     */
    public void attachListener(CcpMessageListener listener) {
        ccpSession.attachListener(listener);
    }

    /**
     * 剔除监听客户端
     * @param listener
     */
    public void detachListener(CcpMessageListener listener) {
        ccpSession.detachListener(listener);
    }

    /**
     * 发送消息，必须连接成功后
     * 
     * 服务器结果将通过异步形式通知给CcpMessageListener
     * 
     * 所有消息都可以通过此方法调用
     * 
     * @param msg
     */
    public void sendMessageNoWait(AbstractMessage msg) throws CcpException {
        try {
            this.ccpSession.sendMessageNoWait(msg);
        } catch (Exception e) {
            throw new CcpException(e.getMessage(), e);
        }

    }
    
    
    /**
     * 发送消息，必须连接成功后
     * 
     * 服务器结果将通过异步形式通知给CcpMessageListener
     * 
     * 所有消息都可以通过此方法调用
     * 
     * autoseq：是否自动生成sequeceId
     * 
     * @param msg
     */
    public void sendMessageNoWait(AbstractMessage msg,boolean autoSeq) throws CcpException {
        try {
            this.ccpSession.sendMessageNoWait(msg,autoSeq);
        } catch (Exception e) {
            throw new CcpException(e.getMessage(), e);
        }

    }

    /**
     * 发送消息，必须连接成功后
     * 并且等待服务器回执,比如发送一个 publish，需要同步知道服务器的ack，那么可以使用此方法
     * 和sendMessageNoWait区别是这个方法会导致当前线程block，直到服务器有对应结果返回
     * timeout为客户端指定超时时间，单位ms，如果超时将抛出异常
     * @param msg
     */
    public SynRespToken sendMessageAndWait(AbstractMessage msg, long timeout) throws CcpException {
        try {
            return this.ccpSession.sendMessageAndWait(msg, timeout);
        } catch (Exception e) {
            throw new CcpException(e.getMessage(), e);
        }

    }

    /**
     * -----------------------
     * -----pub 协议-----------
     * -----------------------
     * 同步调用 等待服务器结果
     * publish and wait
     * 等待timeToWait时间，如果服务器没有puback回执，则认为失败了。
     * @param msg
     * @param timeout
     * @return
     * @throws CcpException
     */
    public PubAck publish(Publish msg) throws CcpException {
        return this.publishAndWait(msg, -1);
    }

    /**
     * -----------------------
     * -----pub 协议-----------
     * -----------------------
     * 同步调用 等待服务器结果
     * publish and wait
     * 等待timeToWait时间，如果服务器没有puback回执，则认为失败了。
     * @param msg
     * @param timeout
     * @return
     * @throws CcpException
     */
    public PubAck publishAndWait(Publish msg, long timeout) throws CcpException {
        try {
            SynRespToken t = this.sendMessageAndWait(msg, timeout);
            if (!t.isError()) {
                return (PubAck) t.getResponse();
            } else {
                throw new CcpException("调用失败: code=" + t.getErrorCode());
            }
        } catch (CcpException e) {
            throw e;
        }
    }

    /** 
     * -----------------------
     * -----RPC 协议-----------
     * -----------------------
     * 同步调用 等待服务器结果
     * 等待timeToWait时间，如果服务器没有Response回执，则认为失败了。
     * @param msg
     * @param timeout
     * @return
     * @throws CcpException
     */
    public Response requestAndWait(Request msg) throws CcpException {
        return requestAndWait(msg, -1);
    }

    /**
     * -----------------------
     * -----RPC 协议-----------
     * -----------------------
     * 同步调用 等待服务器结果
     * 等待timeToWait时间，如果服务器没有Response回执，则认为失败了。
     * @param msg
     * @param timeout
     * @return
     * @throws CcpException
     */
    public Response requestAndWait(Request msg, long timeout) throws CcpException {
        try {
            SynRespToken t = this.sendMessageAndWait(msg, timeout);
            if (!t.isError()) {
                return (Response) t.getResponse();
            } else {
                throw new CcpException("调用失败: code=" + t.getErrorCode());
            }
        } catch (CcpException e) {
            throw e;
        }
    }

    /**
     * 返回当前会话id
     * @return
     */
    public String getSessionId() {
        return ccpSession.getSessionId();
    }
    
    
    /**
     * 返回当前会话token 
     * @return
     */
    public String getToken() {
        return ccpSession.getToken();
    }
    

}
