/**
 * Alibaba.com Inc.
 * Copyright (c) 2004-2016 All Rights Reserved.
 */
package com.aliyun.iot.connect;

import java.io.ByteArrayInputStream;
import java.security.PublicKey;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import com.aliyun.iot.CcpException;
import com.aliyun.iot.util.AliyunWebUtils;
import com.aliyun.iot.util.Base64Util;
import com.aliyun.iot.util.CipherUtils;
import com.aliyun.iot.util.LogUtil;
import com.aliyun.iot.util.SignUtil;

/**
 * 授权认证
 *
 * @version $Id: IotAuthUtil.java,v 0.1 2016年3月30日 下午8:14:38  Exp $
 */
public class Auth {

    /**
     * (1)调用设备配置接口初始化设备，服务器返回公钥以及数据网关服务器地址
     * @param params
     * @throws Exception 
     */
    public static Map<String, String> auth(ConnectOpts opts) throws Exception {

        LogUtil.print("进行设备授权  /iot/auth");
        /*
         * 填充http参数，除了服务器指定的必须字段外
         * 客户端可以透传一些属性以增强sign的健壮性
         * 必传参数：deviceId productKey sign
         */
        Map<String, String> params = new HashMap<String, String>();

        if (opts.getDeviceId() != null) {
            params.put("deviceId", opts.getDeviceId());
        }

        if (opts.getDeviceName() != null) {
            params.put("deviceName", opts.getDeviceName());
        }

        params.put("productKey", opts.getProductAppKey());
        //params.put("appKey", opts.getProductAppKey());
        //params.put("hmac", opts.getSignMethod());
        params.put("signMethod", opts.getSignMethod());
        //params.put("protocol", "ccp");
        //params.put("resFlag", "ip");
        params.put("time", new SimpleDateFormat("yyyy-MM-dd hh:mm:ss").format(new Date()));

        /*
         *生成sign签名，发送给服务器的参数都需要加入sign签名列表
         */
        params.put(
            "sign",
            SignUtil.sign(params, opts.getProductAppSecret(), opts.getDeviceSecret(),
                opts.getSignMethod()));

        //调用配置认证接口，获得seedkey以及ccp连接数据通道地址
        String result = AliyunWebUtils.doPost(opts.getAuthUrl() + "/iot/auth", params, 5000, 5000);
        LogUtil.print("调用设备授权认证接口 response：" + result);

        //解析结果
        Map<String, String> json = result2map(result);
        if (!"true".equals(json.get("success"))) {
            throw new CcpException("设备授权失败");
        }

        return json;

    }

    /**
     * (2)生成您自己的私钥seedkey，并使用服务器公钥加密发给服务器，成功后得到sessionId （sid）
     * @param params
     * @param pubkey
     * @param result
     * @throws Exception 
     */
    public static String getSessionid(ConnectOpts opts, String pubkey, String pkVersion,
                                      byte[] seedkey) throws Exception {

        LogUtil.print("调用密钥交换确认会话接口  /iot/sid");
        //交换秘钥并获得sessionid
        Map<String, String> params2 = new HashMap<String, String>();

        CertificateFactory cf = CertificateFactory.getInstance("X.509");
        X509Certificate cert = (X509Certificate) cf.generateCertificate(new ByteArrayInputStream(
            pubkey.getBytes()));
        PublicKey publicKey = cert.getPublicKey();
        opts.setPkey(publicKey);

        String seedKey = Base64Util.encode(CipherUtils.encryptBytesByPublicKey(seedkey, publicKey));

        //加密后的seedKey
        params2.put("seedKey", seedKey);
        params2.put("pkVersion", pkVersion);//证书版本
        params2.put("signMethod", opts.getSignMethod());
        //params2.put("hmac", Config.hmac);
        String data = "{deviceId:\"" + (opts.getDeviceId() == null ? "" : opts.getDeviceId())
                      + "\", productKey:\"" + opts.getProductAppKey() + "\",deviceName:\""
                      + (opts.getDeviceName() == null ? "" : opts.getDeviceName()) + "\" }";
        //String data = "{deviceId:\"" + opts.getDeviceId() + "\", appKey:\"" + opts.getProductAppKey() + "\"}";

        params2.put("data",
            Base64Util.encode(CipherUtils.encodeDataWithAES(data.getBytes(), seedkey)));

        params2.put(
            "sign",
            SignUtil.sign(params2, opts.getProductAppSecret(), opts.getDeviceSecret(),
                opts.getSignMethod()));

        String result2 = AliyunWebUtils.doPost(opts.getAuthUrl() + "/iot/sid", params2, 5000, 5000);
        LogUtil.print("获取会话认证response：" + result2);
        //解析结果
        Map<String, String> json = result2map(result2);

        if (!"true".equals(json.get("success"))) {
            throw new CcpException("获取sid失败");
        }

        byte[] sidarray = Base64Util.decode(json.get("sid"));
        byte[] sidfinal = CipherUtils.aesDecrypt(sidarray, seedkey);

        String sid = new String(sidfinal);
        LogUtil.print("解密后的sid=" + sid);
        return sid;

    }

    /**
     * 简单的json字符串解析
     * @param result
     * @return
     */
    public static Map<String, String> result2map(String result) {
        String[] jsonarray = result.replace("{", "").replace("}", "").replaceAll("[\\n]", "")
            .split(",");
        Map<String, String> json = new HashMap<String, String>();
        for (String item : jsonarray) {
            String[] keyvalue = item.split("\":");
            json.put(keyvalue[0].replaceAll("\"", "").trim(), keyvalue[1].replaceAll("\"", "")
                .trim());
        }
        return json;
    }

}
