/**
 * Alibaba.com Inc.
 * Copyright (c) 2004-2016 All Rights Reserved.
 */
package com.aliyun.iot.internal;

import java.io.IOException;
import java.net.Socket;
import java.util.Date;
import java.util.Map;
import java.util.Queue;
import java.util.Random;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import com.aliyun.iot.CcpException;
import com.aliyun.iot.CcpMessageListener;
import com.aliyun.iot.connect.Auth;
import com.aliyun.iot.connect.ConnectOpts;
import com.aliyun.iot.handler.CcpMessagehandler;
import com.aliyun.iot.handler.RecieveMessageThread;
import com.aliyun.iot.message.AbstractMessage;
import com.aliyun.iot.message.ConnectACK;
import com.aliyun.iot.message.ConnectExtAuth;
import com.aliyun.iot.message.MessageHeader;
import com.aliyun.iot.message.MessagePacket;
import com.aliyun.iot.message.PingRequestACK;
import com.aliyun.iot.message.PubAck;
import com.aliyun.iot.message.Publish;
import com.aliyun.iot.message.Push;
import com.aliyun.iot.message.PushACK;
import com.aliyun.iot.message.Reconnect;
import com.aliyun.iot.message.ReconnectACK;
import com.aliyun.iot.message.Request;
import com.aliyun.iot.message.Response;
import com.aliyun.iot.message.StatusCode;
import com.aliyun.iot.message.UnknownSession;
import com.aliyun.iot.util.Base64Util;
import com.aliyun.iot.util.CipherUtils;
import com.aliyun.iot.util.LogUtil;

/**
 * ccp 会话
 * <p>具体说明</p>
 *
 * @version $Id: Session.java,v 0.1 2016年6月28日 下午3:17:06  Exp $
 */
public class Session {

    /** 会话状态 **/
    private StatusCode                statusCode;

    /** 会话token */
    private String                    token;

    /** seedkey **/
    private byte[]                    seedKey;

    /** 阿里云设备对应的全局设备id **/
    private String                    deviceId;

    /** 当前证书版本 **/
    private String                    pkVersion;

    /** crt证书字符串 **/
    private String                    pubkeyCrt;

    /** socket */
    private Socket                    socket;

    /** sid **/
    private String                    sessionId;

    /** 消息通道 **/
    private String                    serverList;

    /** 本地消息sequenceId 计数器 **/
    private AtomicInteger             sequenceId      = new AtomicInteger(1000);

    /** 消息客户端监听者 **/
    private Queue<CcpMessageListener> listeners       = new ConcurrentLinkedQueue<CcpMessageListener>();

    /** 消息处理器 **/
    private CcpMessagehandler         handler         = null;

    private static final int          CORE_POOL_SIZE  = Runtime.getRuntime().availableProcessors() <= 1 ? 1
                                                          : Runtime.getRuntime()
                                                              .availableProcessors();

    final static ExecutorService      roamxecutor     = new ThreadPoolExecutor(CORE_POOL_SIZE, 50,
                                                          10, TimeUnit.SECONDS,
                                                          new ArrayBlockingQueue<Runnable>(10));

    /** timer **/
    private Timer                     timer           = null;

    /** connect opts **/
    private ConnectOpts               connectOpts     = null;

    /** 上一次收发消息时间 **/
    private Date                      lastMessageTime = new Date();

    private RecieveMessageThread      recv            = null;

    /** token store **/
    private TokenStore                tokenStore      = null;

    /**
     * construct
     */
    public Session(ConnectOpts connectOpts) {
        this.connectOpts = connectOpts;

        tokenStore = new TokenStore();
    }

    /**
     * 生成随机的私钥
     * @return
     */
    private byte[] generateRamdomSeedKey() {
        byte[] seedkey = new byte[16];
        new Random().nextBytes(seedkey);
        return seedkey;
    }

    /**
     * 授权操作，不用每次都调用
     * sid有效期 24小时
     * @throws Exception
     */
    public synchronized void auth() throws CcpException {

        try {
            /** (1)调用设备配置接口初始化设备，服务器返回公钥以及数据通道服务器地址 **/
            Map<String, String> result = Auth.auth(connectOpts);
            this.setPkVersion(result.get("pkVersion"));
            this.setServerList(result.get("servers"));
            this.deviceId = result.get("deviceId");
            String pubkey = new String(Base64Util.decode(result.get("pubkey")), "utf-8");
            this.pubkeyCrt = pubkey;

            /** (2)生成您自己的私钥seedkey，并使用服务器公钥加密发给服务器，成功后得到sessionId （sid）**/
            byte[] seedkey = generateRamdomSeedKey();
            LogUtil.print("seedkey byte2hexstring:" + CipherUtils.bytesToHexString(seedkey));
            this.setSeedKey(seedkey);

            String sid = Auth.getSessionid(connectOpts, this.pubkeyCrt, this.pkVersion,
                this.seedKey);
            /** (3)使用sid连接数据通道（即第一步返回的服务器地址），进行数据传输 **/
            this.setSessionId(sid);

            // init timer for reconnect, and update session id, as it will expired in 24 hours
            if (timer == null) {
                timer = new Timer("timmer-" + this.getSessionId() + "-"
                                  + System.currentTimeMillis());
            } else {
                timer.cancel();
                timer = new Timer("timmer-" + this.getSessionId() + "-"
                                  + System.currentTimeMillis());
            }
            //            if (authTask == null) {
            //                authTask = new ReauthTask();
            //                //23小时自动刷新sid
            //                timer.scheduleAtFixedRate(authTask, 23 * 60 * 60 * 1000, 23 * 60 * 60 * 1000);
            //            }
        } catch (Exception e) {
            throw new CcpException(e.getMessage(), e);
        }
    }

    /**
     * 有消息到达
     * 
     * @param header
     * @param payload
     * @throws Exception
     */
    public void onMessage(MessageHeader header, byte[] payload) throws Exception {
        //更新时间戳
        reschedulePing();

        AbstractMessage msg = handler.hander(header, payload);

        SynRespToken tk = tokenStore.getToken(msg);
        if (tk != null) {
            //如果有token，说明是同步方式调用，无需异步通知
            tk.onMessageCheck(msg);
            tokenStore.removeToken(msg);
        } else {
            //异步方式通知
            if (header.messageType != MessagePacket.CONNECT_ACK
                && header.messageType != MessagePacket.RECONNECT_ACK) {
                //非连接消息
                this.notifyMsg(msg);
            }
        }

        // ----以下部分协议部分需要保持心跳和回执------
        if (header.messageType == MessagePacket.CONNECT_ACK) {
            ConnectACK m = (ConnectACK) msg;
            // 提取token到session
            this.setToken(m.connectionToken);
            this.setStatusCode(m.statusCode);
            this.connectOpts.setAliveSeconds(m.keepAlive);
            //通知观察者
            if (m.statusCode == StatusCode.success) {
                this.notifyConnect(m);
                scheduleReconnectCancel();
                return;
            }
        } else if (header.messageType == MessagePacket.RECONNECT_ACK) {
            ReconnectACK m = (ReconnectACK) msg;
            // 提取token到session
            this.setToken(m.connectionToken);
            this.setStatusCode(m.statusCode);
            this.connectOpts.setAliveSeconds(m.keepAlive);

            ConnectACK m2 = new ConnectACK();
            m2.statusCode = m.statusCode;
            m2.connectionToken = m.connectionToken;
            m2.keepAlive = m.keepAlive;

            if (m.statusCode == StatusCode.success) {
                this.notifyConnect(m2);
                scheduleReconnectCancel();
                return;
            }
        } else if (header.messageType == MessagePacket.PINGREQUEST) {
            // 心跳包 返回
            this.sendMessageNoWait(new PingRequestACK());
        } else if (header.messageType == MessagePacket.PUBLISH) {
            // 对方推送的消息
            Publish m = (Publish) msg;
            if (m.aliveSecond > 0) {
                // 需要回执
                PubAck ack = new PubAck();
                ack.sequenceID = m.sequenceID;
                this.sendMessageNoWait(ack,false);
            }

        } 
//rrpc的response应该由设备端来回复内容，而不是sdk自动回复 20160918 
//        else if (header.messageType == MessagePacket.R_REQUEST) {
//            // 需要回执
//            RResponse mb = new RResponse();
//            mb.sequenceID = msg.sequenceID;
//            mb.payload = "ok".getBytes();
//            this.sendMessageNoWait(mb,false);
//        }
        else if (header.messageType == MessagePacket.PUSH) {
            // 需要回执
            Push m = (Push) msg;
            PushACK mb = new PushACK();
            mb.sequenceID = msg.sequenceID;
            mb.messageID = m.messageID;
            mb.appID = m.appID;
            this.sendMessageNoWait(mb,false);
        } else if (header.messageType == MessagePacket.UNKNOWN_SESSION) {
            //unknownsession 是所有错误的消息需要 细化里面的类型，如果证书过期，自动刷新
            UnknownSession err = (UnknownSession) msg;
            handlerUnknowSessionError(err);
        }

    }

    /**
     * 刷新心跳时间
     */
    private synchronized void reschedulePing() {
        lastMessageTime = new Date();
        if (checkPingTask == null) {
            checkPingTask = new CheckPingTask();
            //检查 ping pong
            timer.schedule(checkPingTask, connectOpts.getAliveSeconds() * 1000,
                connectOpts.getAliveSeconds() * 1000);
        } else {
            checkPingTask.cancel();
            checkPingTask = new CheckPingTask();
            timer.schedule(checkPingTask, connectOpts.getAliveSeconds() * 1000,
                connectOpts.getAliveSeconds() * 1000);
        }
    }

    /**
     * 自动处理可预知的错误
     * @param err
     */
    private void handlerUnknowSessionError(UnknownSession err) {
        switch (err.statusCode) {
            case reject_ca_expired:
                this.sessionId = null;
                //证书过期
                LogUtil.print("证书过期，需要重连接");
                statusCode = StatusCode.offline;
                this.connect();
                break;
            case reject_token_expired:
            case reject_invalid_token:
                //this.sessionId = null;
                //token 过期，重新连接 ,或者token无效 ，可以重新连接获取
                LogUtil.print("token过期，需要重连接");
                statusCode = StatusCode.offline;
                this.connect();
                break;
            case reject_invalid_sid:
                try {
                    String sid = Auth.getSessionid(connectOpts, this.pubkeyCrt, this.pkVersion,
                        this.seedKey);
                    /** 重新设置sid **/
                    this.setSessionId(sid);
                } catch (Exception e) {
                    e.printStackTrace();
                    //无法修复，重新连接
                    this.sessionId = null;
                }
                statusCode = StatusCode.offline;
                this.connect();
                break;
            default:
                //nothing
        }
    }

    /**
     * 增加监听客户端
     * @param listener
     */
    public void attachListener(CcpMessageListener listener) {
        listeners.add(listener);
    }

    /**
     * 剔除监听客户端
     * @param listener
     */
    public void detachListener(CcpMessageListener listener) {
        listeners.remove(listener);
    }

    /**
     * 通知观察者
     * @param msg
     */
    private synchronized void notifyMsg(final AbstractMessage msg) {
        for (final CcpMessageListener lister : listeners) {

            roamxecutor.execute(new Runnable() {
                @Override
                public void run() {
                    lister.onMessage(msg);
                }
            });

        }
    }

    /**
     * 通知观察者
     * @param msg
     */
    private synchronized void notifyConnect(final ConnectACK msg) {
        for (final CcpMessageListener lister : listeners) {

            roamxecutor.execute(new Runnable() {
                @Override
                public void run() {
                    lister.onConnect(msg);
                }
            });

        }
    }

    /**
     * 通知观察者
     * @param msg
     */
    public synchronized void notifyOffline() {
        for (final CcpMessageListener lister : listeners) {

            roamxecutor.execute(new Runnable() {
                @Override
                public void run() {
                    lister.onOffLine();
                }
            });

            // 自动尝试重连 
            if (this.connectOpts.isAutoReconnect()) {
                LogUtil.print("auto reconnect timmer for sid: " + this.getSessionId());
                if (reconnectTask == null) {
                    reconnectTask = new ReConnectTask();
                    timer.schedule(reconnectTask, 0, this.connectOpts.getAutoReconnectDelay());
                }
            }
        }
    }

    private synchronized void scheduleReconnectCancel() {
        if (reconnectTask != null) {
            reconnectTask.cancel();
            reconnectTask = null;
        }
    }

    private ReConnectTask reconnectTask = null;

    private class ReConnectTask extends TimerTask {
        public void run() {
            LogUtil.print("ReConnectTask--" + Thread.currentThread().getId() + " -- "
                          + Thread.currentThread().getName());
            reConnect();
        }
    }

    private CheckPingTask checkPingTask = null;

    private class CheckPingTask extends TimerTask {
        public void run() {
            LogUtil.print("CheckPingTask t--" + Thread.currentThread().getId() + " -- "
                          + Thread.currentThread().getName());
            if (statusCode == StatusCode.connecting) {
                return;
            }
            long delay = System.currentTimeMillis() - lastMessageTime.getTime();
            //允许误差2秒
            if (delay > (connectOpts.getAliveSeconds() * 1000 + 2000)) {
                LogUtil.print("CheckPingTask delay-- offline");
                statusCode = StatusCode.offline;
                notifyOffline();
            }
        }
    }

    //private ReauthTask authTask = null;

    //    private class ReauthTask extends TimerTask {
    //
    //        public void run() {
    //            LogUtil.print("ReauthTask 开始刷新sid");
    //            String sid;
    //            try {
    //                sid = Auth.getSessionid(connectOpts, getPubkeyCrt(), getPkVersion(), getSeedKey());
    //                setSessionId(sid);
    //                statusCode = StatusCode.offline;
    //                notifyOffline();
    //            } catch (Exception e) {
    //                throw new CcpException(e.getMessage(), e);
    //            }
    //
    //        }
    //    }

    /**
     * 同步调用 等待服务器结果
     * publish and wait
     * 等待timeToWait时间，如果服务器没有puback回执，则认为失败了。
     * @param msg
     * @param timeout
     * @return
     * @throws CcpException
     */
    public PubAck publish(Publish msg) throws CcpException {
        return this.publishAndWait(msg, -1);
    }

    /**
     * 同步调用 等待服务器结果
     * publish and wait
     * 等待timeToWait时间，如果服务器没有puback回执，则认为失败了。
     * @param msg
     * @param timeout
     * @return
     * @throws CcpException
     */
    public PubAck publishAndWait(Publish msg, long timeout) throws CcpException {
        try {
            SynRespToken t = this.sendMessageAndWait(msg, timeout);
            if (!t.isError()) {
                return (PubAck) t.getResponse();
            } else {
                throw new CcpException("调用失败: code=" + t.getErrorCode());
            }
        } catch (CcpException e) {
            throw e;
        }
    }

    /**
     * 同步调用 等待服务器结果
     * 等待timeToWait时间，如果服务器没有Response回执，则认为失败了。
     * @param msg
     * @param timeout
     * @return
     * @throws CcpException
     */
    public Response requestAndWait(Request msg) throws CcpException {
        return requestAndWait(msg, -1);
    }

    /**
     * 同步调用 等待服务器结果
     * 等待timeToWait时间，如果服务器没有Response回执，则认为失败了。
     * @param msg
     * @param timeout
     * @return
     * @throws CcpException
     */
    public Response requestAndWait(Request msg, long timeout) throws CcpException {
        try {
            SynRespToken t = this.sendMessageAndWait(msg, timeout);
            if (!t.isError()) {
                return (Response) t.getResponse();
            } else {
                throw new CcpException("调用失败: code=" + t.getErrorCode());
            }
        } catch (CcpException e) {
            throw e;
        }
    }

    /**
     * 
     * 发送消息，必须连接成功后
     * 异步等待消息
     * @param msg
     */
    public void sendMessageNoWait(AbstractMessage msg) {
        this.sendMessageNoWait(msg, true);
    }
    
    
    /**
     * 
     * 发送消息，必须连接成功后
     * 异步等待消息
     * @param msg
     */
    public void sendMessageNoWait(AbstractMessage msg, boolean autoSeq) {

        try {

            if (this.statusCode != StatusCode.success) {
                LogUtil.print("发送消息网络未连接");
                reConnect();// 尝试连接
                return;
            } else {
                //消息id 自增
                if(autoSeq){
                    msg.sequenceID = sequenceId.incrementAndGet();
                }
                byte[] data = msg.encode(this.seedKey);
                socket.getOutputStream().write(data);
                socket.getOutputStream().flush();
                LogUtil.print("消息发送成功:" + msg);
            }
        } catch (IOException e) {
            this.statusCode = StatusCode.offline;
            this.notifyOffline();
            throw new CcpException(e.getMessage(), e);
        } catch (Exception e) {
            throw new CcpException(e.getMessage(), e);
        }

    }

    /**
     * 发送消息，必须连接成功后
     * 并且等待服务器回执,比如发送一个 publish，需要同步知道服务器的ack，那么可以使用此方法
     * timeout为指定超时时间，单位ms
     * @param msg
     */
    public SynRespToken sendMessageAndWait(AbstractMessage msg, long timeout) {

        try {

            if (this.statusCode != StatusCode.success) {
                LogUtil.print("发送消息网络未连接");
                throw new CcpException("发送消息网络未连接");
            } else {
                //消息id 自增
                msg.sequenceID = sequenceId.incrementAndGet();
                SynRespToken t = tokenStore.createToken(msg);
                byte[] data = msg.encode(this.seedKey);
                socket.getOutputStream().write(data);
                socket.getOutputStream().flush();
                LogUtil.print("消息发送成功，等待回执消息, id=" + msg.sequenceID);
                t.waitForComplete(timeout);
                return t;
            }
        } catch (IOException e) {
            this.statusCode = StatusCode.offline;
            this.notifyOffline();
            throw new CcpException(e.getMessage(), e);
        } catch (Exception e) {
            throw new CcpException(e.getMessage(), e);
        }

    }

    /**
     * 快速重连
     * 针对已有sid和token情况
     * @throws Exception
     */
    private synchronized void reConnect() {

        try {
            if (this.statusCode == StatusCode.connecting || this.statusCode == StatusCode.success) {
                return;
            }

            tokenStore.shutDown();

            this.statusCode = StatusCode.connecting;
            //sequenceId.set(0);
            // tcp连接网关数据通信
            String[] host = serverList.split(":");
            String[] ports = host[1].split("\\|");
            socket = new Socket(host[0], Integer.parseInt(ports[0]));
            socket.setKeepAlive(true);
            socket.setSoTimeout(connectOpts.getAliveSeconds() * 1000);

            Reconnect msg = new Reconnect(this.connectOpts);
            msg.deviceID = this.deviceId;
            msg.keepAlive = connectOpts.getAliveSeconds();
            msg.connectionToken = this.getToken();
            msg.encryptPayload(this.getSeedKey());
            socket.getOutputStream().write(msg.encode(seedKey));
            socket.getOutputStream().flush();
            new RecieveMessageThread(socket, this).start();
        } catch (Exception e) {
            this.statusCode = StatusCode.offline;
            e.printStackTrace();
        }

    }

    /**
     * 连接初始化
     * 
     * @throws Exception
     */
    public synchronized void connect() {
        try {

            if (this.statusCode == StatusCode.connecting) {
                throw new CcpException("正在连接中...");
            }
            this.statusCode = StatusCode.connecting;
            //sequenceId.set(0);

            if (recv != null && !recv.isStop()) {
                recv.setStop(true);
            }

            if (seedKey == null || this.sessionId == null || this.serverList == null) {
                //如果还未授权，则先授权
                auth();
            }

            if (handler == null) {
                // 监听通道读取消息
                handler = new CcpMessagehandler(seedKey);
            } else {
                handler.setSeedKey(seedKey);
            }

            tokenStore.shutDown();

            // tcp连接网关数据通信
            String[] host = serverList.split(":");
            String[] ports = host[1].split("\\|");
            socket = new Socket(host[0], Integer.parseInt(ports[0]));
            socket.setKeepAlive(true);
            socket.setSoTimeout(connectOpts.getAliveSeconds() * 1000);

            ConnectExtAuth connectmessage = new ConnectExtAuth();
            connectmessage.sid = sessionId;
            connectmessage.platformId = 2;
            connectmessage.keepAlive = connectOpts.getAliveSeconds();
            connectmessage.sequenceID = sequenceId.incrementAndGet();
            connectmessage.encryptPayload(seedKey);
            socket.getOutputStream().write(connectmessage.encode(seedKey));
            socket.getOutputStream().flush();
            recv = new RecieveMessageThread(socket, this);
            recv.start();
        } catch (Exception e) {
            this.statusCode = StatusCode.offline;
            throw new CcpException("连接失败：" + e.getMessage(), e);
        }
    }

    /**
     * Getter method for property <tt>statusCode</tt>.
     * 
     * @return property value of statusCode
     */
    public StatusCode getStatusCode() {
        return statusCode;
    }

    /**
     * Setter method for property <tt>statusCode</tt>.
     * 
     * @param statusCode
     *            value to be assigned to property statusCode
     */
    public void setStatusCode(StatusCode statusCode) {
        this.statusCode = statusCode;
    }

    /**
     * Getter method for property <tt>token</tt>.
     * 
     * @return property value of token
     */
    public String getToken() {
        return token;
    }

    /**
     * Setter method for property <tt>token</tt>.
     * 
     * @param token
     *            value to be assigned to property token
     */
    public void setToken(String token) {
        this.token = token;
    }

    /**
     * Getter method for property <tt>socket</tt>.
     * 
     * @return property value of socket
     */
    public Socket getSocket() {
        return socket;
    }

    /**
     * Setter method for property <tt>socket</tt>.
     * 
     * @param socket
     *            value to be assigned to property socket
     */
    public void setSocket(Socket socket) {
        this.socket = socket;
    }

    /**
     * Getter method for property <tt>seedKey</tt>.
     * 
     * @return property value of seedKey
     */
    public byte[] getSeedKey() {
        return seedKey;
    }

    /**
     * Setter method for property <tt>seedKey</tt>.
     * 
     * @param seedKey
     *            value to be assigned to property seedKey
     */
    public void setSeedKey(byte[] seedKey) {
        this.seedKey = seedKey;
    }

    /**
     * Getter method for property <tt>sessionId</tt>.
     * 
     * @return property value of sessionId
     */
    public String getSessionId() {
        return sessionId;
    }

    /**
     * Setter method for property <tt>sessionId</tt>.
     * 
     * @param sessionId
     *            value to be assigned to property sessionId
     */
    public void setSessionId(String sessionId) {
        this.sessionId = sessionId;
    }

    /**
     * Getter method for property <tt>serverList</tt>.
     * 
     * @return property value of serverList
     */
    public String getServerList() {
        return serverList;
    }

    /**
     * Setter method for property <tt>serverList</tt>.
     * 
     * @param serverList value to be assigned to property serverList
     */
    public void setServerList(String serverList) {
        this.serverList = serverList;
    }

    /**
     * Getter method for property <tt>pkVersion</tt>.
     * 
     * @return property value of pkVersion
     */
    public String getPkVersion() {
        return pkVersion;
    }

    /**
     * Setter method for property <tt>pkVersion</tt>.
     * 
     * @param pkVersion value to be assigned to property pkVersion
     */
    public void setPkVersion(String pkVersion) {
        this.pkVersion = pkVersion;
    }

    /**
     * Getter method for property <tt>pubkeyCrt</tt>.
     * 
     * @return property value of pubkeyCrt
     */
    public String getPubkeyCrt() {
        return pubkeyCrt;
    }

    /**
     * Setter method for property <tt>pubkeyCrt</tt>.
     * 
     * @param pubkeyCrt value to be assigned to property pubkeyCrt
     */
    public void setPubkeyCrt(String pubkeyCrt) {
        this.pubkeyCrt = pubkeyCrt;
    }

    /**
     * Getter method for property <tt>connectOpts</tt>.
     * 
     * @return property value of connectOpts
     */
    public ConnectOpts getConnectOpts() {
        return connectOpts;
    }

    /**
     * Setter method for property <tt>connectOpts</tt>.
     * 
     * @param connectOpts value to be assigned to property connectOpts
     */
    public void setConnectOpts(ConnectOpts connectOpts) {
        this.connectOpts = connectOpts;
    }
}
