/**
 * Alibaba.com Inc.
 * Copyright (c) 2004-2016 All Rights Reserved.
 */
package com.aliyun.iot.internal;

import com.aliyun.iot.CcpException;
import com.aliyun.iot.message.AbstractMessage;
import com.aliyun.iot.message.ConnectACK;
import com.aliyun.iot.message.MessagePacket;
import com.aliyun.iot.message.PubAck;
import com.aliyun.iot.message.ReconnectACK;
import com.aliyun.iot.message.StatusCode;
import com.aliyun.iot.util.LogUtil;

/**
 * 同步调用器
 * token for lock
 * @version $Id: Token.java,v 0.1 2016年6月28日 上午10:13:10  Exp $
 */
public class SynRespToken {

    /** uni key 对应某一个消息 sequenceId **/
    private String          key;

    /** 是否有错误 **/
    private boolean         isError;

    /** 错误码 请参考 statusCode **/
    private int             errorCode = 0;

    /** 返回消息体 **/
    private AbstractMessage response;

    /** 是否已完成 **/
    private boolean         isComplete;

    private Object          lock      = new Object();

    public SynRespToken(String key) {
        this.key = key;
        isComplete = false;
        LogUtil.print("token create :" + key);
    }

    /**
     * wait no timeout
     * 
     * @throws CcpException
     */
    public void waitForComplete() throws CcpException {
        this.waitForComplete(-1);
    }

    /**
     * wait with timeout
     * 
     * @param timeout
     * @throws CcpException
     */
    public void waitForComplete(long timeout) throws CcpException {
        try {
            if (response != null) {
                return;
            }
            synchronized (lock) {
                if (timeout > 0) {
                    lock.wait(timeout);
                    if (this.response == null) {
                        throw new CcpException("wait timeout for:" + key);
                    }
                } else {
                    lock.wait();
                }
            }
        } catch (Exception e) {
            throw new CcpException(e.getMessage(), e);
        } finally {
            isComplete = true;
        }

    }

    public static String generateKey(AbstractMessage message) {
        return "seq_" + message.sequenceID;
    }

    /**
     * check if lock complete
     * 
     * @param message
     */
    public void onMessageCheck(AbstractMessage message) {
        if (key.equals(generateKey(message))) {
            //LogUtil.print("onMessageCheck -- notifyComplete:" + key);
            this.response = message;
            this.isError = false;
            if (message.getMessageType() == MessagePacket.PUSH_ACK) {
                PubAck msg = (PubAck) message;
                this.isError = (msg.code != StatusCode.success.code());
                this.errorCode = msg.code;
            } else if (message.getMessageType() == MessagePacket.CONNECT_ACK) {
                ConnectACK msg = (ConnectACK) message;
                this.isError = (msg.statusCode != StatusCode.success);
                this.errorCode = msg.statusCode.code();
            } else if (message.getMessageType() == MessagePacket.RECONNECT_ACK) {
                ReconnectACK msg = (ReconnectACK) message;
                this.isError = (msg.statusCode != StatusCode.success);
                this.errorCode = msg.statusCode.code();
            } else if (message.getMessageType() == MessagePacket.UNKNOWN_SESSION) {
                this.isError = true;
            }
            notifyComplete(this.isError);
        } else {
            LogUtil.print("onMessageCheck -- ingore");
        }
    }

    /**
     * complete
     */
    public void notifyComplete(boolean isError) {
        synchronized (lock) {
            //LogUtil.print("token notifyComplete :" + key);
            this.isError = isError;
            lock.notifyAll();
        }
    }

    /**
     * Getter method for property <tt>isError</tt>.
     * 
     * @return property value of isError
     */
    public boolean isError() {
        return isError;
    }

    /**
     * Getter method for property <tt>response</tt>.
     * 
     * @return property value of response
     */
    public AbstractMessage getResponse() {
        return response;
    }

    /**
     * Getter method for property <tt>isComplete</tt>.
     * 
     * @return property value of isComplete
     */
    public boolean isComplete() {
        return isComplete;
    }

    /**
     * Getter method for property <tt>errorCode</tt>.
     * 
     * @return property value of errorCode
     */
    public int getErrorCode() {
        return errorCode;
    }

}
