package com.aliyun.iot.message;

import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.spec.InvalidKeySpecException;

import javax.crypto.BadPaddingException;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;

import com.aliyun.iot.util.CipherUtils;
import com.aliyun.iot.util.DynamicByteBuffer;
import com.aliyun.iot.util.EncryptType;

/**
 * 所有消息的基类，只包含固定头相关的元信息。
 */
public abstract class AbstractMessage implements MessagePacket {

    protected MessageHeader fixedHeader;

    public int              sequenceID = 0; // 请求序号

    public AbstractMessage(int messageType) {

        this(new MessageHeader(messageType));
    }

    public AbstractMessage(MessageHeader fixedHeader) {

        this.fixedHeader = fixedHeader;
    }

    /**
     * 封装消息包
     * @param seedKey
     * @return
     */
    public byte[] encode(byte[] seedKey) {

        DynamicByteBuffer result = DynamicByteBuffer.allocate(300);

        //1,判断是否加密
        byte[] bd = encodeBody();
        if (getEncryptType() == EncryptType.symmetrical && bd != null && bd.length > 0) {
            try {
                byte[] data = CipherUtils.aesEncrypt(bd, seedKey);
                this.fixedHeader.remainLength = data.length;
                result.put(this.fixedHeader.generateFixedHeader());
                result.put(data);
            } catch (Exception e) {
                e.printStackTrace();
                return null;
            }
        } else {
            byte[] data = bd;
            if (data != null && data.length > 0) {
                this.fixedHeader.remainLength = data.length;
            }
            result.put(this.fixedHeader.generateFixedHeader());
            if (data != null && data.length > 0) {
                result.put(data);
            }
        }

        return result.array();

    }

    @Override
    public abstract byte[] encodeBody();

    @Override
    public abstract void decodeBody(DynamicByteBuffer buffer);

    /**
     * 解析消息
     * @param payload
     * @throws BadPaddingException 
     * @throws IllegalBlockSizeException 
     * @throws NoSuchPaddingException 
     * @throws NoSuchAlgorithmException 
     * @throws InvalidKeySpecException 
     * @throws InvalidKeyException 
     */
    public void decode(byte[] payload, byte[] seedKey) {

        DynamicByteBuffer buffer = new DynamicByteBuffer(payload);

        //1,判断是否加密
        if (getEncryptType() == EncryptType.symmetrical && payload != null && payload.length > 0) {
            try {
                byte[] data = CipherUtils.aesDecrypt(payload, seedKey);
                buffer = new DynamicByteBuffer(data);
            } catch (Exception e) {
                e.printStackTrace();
                return;
            }
        }

        //2,解析body
        decodeBody(buffer);
    }

    public long getRemainLength() {
        return this.fixedHeader.remainLength;
    }

    public boolean isLastFrame() {
        return this.fixedHeader.compress;
    }

    public void setLastFrame(boolean lastFrame) {
        this.fixedHeader.compress = lastFrame;
    }

    public int getQosLevel() {
        return this.fixedHeader.qosLevel;
    }

    public void setQosLevel(int qosLevel) {
        this.fixedHeader.qosLevel = qosLevel;
    }

    public boolean hasData() {
        return this.fixedHeader.hasData;
    }

    public void setHasData(boolean hasData) {
        this.fixedHeader.hasData = hasData;
    }

    public int getMessageType() {
        return this.fixedHeader.messageType;
    }

    public MessageHeader getFixedHeader() {
        return fixedHeader;
    }

    /**
     * 加密类型
     * @see com.aliyun.iot.message.MessagePacket#getEncryptType()
     */
    public EncryptType getEncryptType() {

        return EncryptType.symmetrical;
    }

    @Override
    public String toString() {
        return getClass().getSimpleName() + "{" + fixedHeader + ", sequenceID=" + sequenceID + '}';
    }

}
